import re

from pydash import get, uniq

from power_platform_security_assessment.base_classes import (
    User, Application, CloudFlow, ConnectorWithConnections, ModelDrivenApp, DesktopFlow
)

_ENV_ID_PATTERN = r"/environments/([^/]+)"


def extract_environment_id(resource_id: str) -> str:
    match = re.search(_ENV_ID_PATTERN, resource_id)
    return match.group(1)


def extract_environment_ids_from_connectors(connectors: list[ConnectorWithConnections]) -> list[str]:
    return uniq(
        extract_environment_id(connection.id)
        for connector in connectors
        for connection in connector.connections
    )


def extract_user_domain(user: User) -> str:
    return user.domainname.split("@")[1].split(".")[0] if '@' in user.domainname else user.domainname


def get_application_owner_id(app: Application) -> str:
    return get(app, 'properties.owner.id')


def get_cloud_flow_owner_id(cloud_flow: CloudFlow) -> str:
    return get(cloud_flow, 'properties.creator.userId')


def get_desktop_flow_owner_id(desktop_flow: DesktopFlow) -> str:
    return desktop_flow.workflowidunique


def get_model_driven_app_owner_id(model_driven_app: ModelDrivenApp) -> str:
    return model_driven_app.appmoduleidunique


def is_app_disabled(app: Application) -> bool:
    return (get(app, 'properties.executionRestrictions.appQuarantineState.quarantineStatus') == 'Quarantined'
            or len(get(app, 'properties.executionRestrictions.dataLossPreventionEvaluationResult.violations', [])) > 0)


def is_flow_disabled(flow: CloudFlow) -> bool:
    return flow.properties.state in ['Stopped', 'Suspended']


def is_model_driven_app_disabled(model_driven_app: ModelDrivenApp) -> bool:
    return model_driven_app.statecode == 1


def is_desktop_flow_disabled(desktop_flow: DesktopFlow) -> bool:
    return desktop_flow.statecode == 1



from pydash import get

from power_platform_security_assessment.base_classes import Application
from power_platform_security_assessment.security_features.bypass_consent.bypass_concent_textual_report import BypassConsentTextualReport
from power_platform_security_assessment.security_features.bypass_consent.model import BypassConsentResult


class BypassConsentAnalyzer:
    def __init__(self, applications: list[Application]):
        self._applications = applications
        self._bypass_consent_textual_report_generator = BypassConsentTextualReport()

    def analyze(self) -> BypassConsentResult:
        bypass_consent_apps = [
            app for app in self._applications
            if app.properties.bypassConsent
               and get(app, 'properties.embeddedApp.type') == 'SharepointFormApp'
        ]

        textual_report = self._bypass_consent_textual_report_generator.generate_textual_report(bypass_consent_apps)

        return BypassConsentResult(
            bypass_consent_apps=bypass_consent_apps,
            textual_report=textual_report,
        )



from urllib.parse import urlencode

from power_platform_security_assessment.base_classes import Application
from power_platform_security_assessment.consts import Requests
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class ApplicationsFetcher(BaseResourceFetcher):
    def __init__(self, env_id: str, token_manager: TokenManager):
        super().__init__(env_id, token_manager)

    def _get_request_url(self):
        environment_id_with_dot = f'{self._env_id[:-2]}.{self._env_id[-2:]}'
        without_dash = environment_id_with_dot.replace('-', '')
        params = {
            "$select": (
                "name,logicalName,properties.displayName,properties.bypassConsent,properties.owner.id,properties.createdBy.id"
                ",properties.executionRestrictions.dataLossPreventionEvaluationResult.violations"
                ",properties.executionRestrictions.appQuarantineState.quarantineStatus"
                ",properties.embeddedApp.type"
            ),
            "api-version": "1",
        }
        return f'https://{without_dash}.environment.api.powerplatform.com/powerapps/apps?{urlencode(params)}'

    def _fetch_single_page_apps(self, token: str, url: str):
        response_data = self._fetch_single_page(url, headers={
            'Authorization': f'Bearer {token}',
            'x-ms-api-context-type': 'Admin',
            'Accept-Encoding': 'gzip,deflate,compress',
        })

        applications = [Application(**app) for app in response_data.get('value', [])]
        next_page = response_data.get('nextLink', None)

        return applications, next_page

    def _fetch_canvas_apps(self) -> list[Application]:
        token = self._token_manager.fetch_access_token(Requests.APPLICATIONS_SCOPE)
        next_page_url = self._get_request_url()
        all_apps = []

        while next_page_url:
            applications, next_page_url = self._fetch_single_page_apps(token=token, url=next_page_url)
            all_apps.extend(applications)
            self._resource_count += len(applications)

        return all_apps

    def _do_fetch_resource_data(self) -> list[Application]:
        return self._fetch_canvas_apps()



from urllib.parse import urlencode

from pydash import map_, filter_

from power_platform_security_assessment.base_classes import Connector, Connection, ConnectorWithConnections
from power_platform_security_assessment.consts import Requests
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class _ConnectionWithConnectorName(Connection):
    connector_name: str


class ConnectionsFetcher(BaseResourceFetcher):
    _CONNECTORS_URL = 'https://api.powerapps.com/api/invoke'
    _IGNORED_CONNECTORS = ['shared_logicflows']

    def __init__(self, env_id: str, token_manager: TokenManager):
        super().__init__(env_id, token_manager)

    def _get_all_connections_url(self):
        params = {"api-version": "2016-11-01"}
        return f'https://api.powerapps.com/providers/Microsoft.PowerApps/scopes/admin/environments/{self._env_id}/connections?{urlencode(params)}'

    def _fetch_connectors(self, token: str, connector_names: list[str]) -> list[Connector]:
        response_data = self._fetch_single_page(
            url=self._CONNECTORS_URL,
            headers={
                'x-ms-path-query': f'/providers/Microsoft.PowerApps/apis?showApisWithToS=true&$filter=environment eq \'{self._env_id}\'&api-version=2020-06-01',
                'Authorization': f'Bearer {token}',
            }
        )
        connectors = [Connector(**connector) for connector in response_data.get('value', [])]
        response_data = self._fetch_single_page(
            url=f'https://api.powerapps.com/providers/Microsoft.PowerApps/scopes/admin/environments/{self._env_id}/apis?api-version=2016-11-01',
            headers={
                'Authorization': f'Bearer {token}',
            }
        )
        custom_connectors = [Connector(**connector) for connector in response_data.get('value', [])]

        additional_used_connectors = []
        for connector_name in connector_names:
            if connector_name not in {connector.name for connector in connectors + custom_connectors}:
                connector_data = self._fetch_single_page(
                    url=self._CONNECTORS_URL,
                    headers={
                        'x-ms-path-query': f'/providers/Microsoft.PowerApps/apis/{connector_name}?showApisWithToS=true&$expand=permissions&$filter=environment eq \'{self._env_id}\'&api-version=2020-06-01',
                        'Authorization': f'Bearer {token}',
                    }
                )
                additional_used_connectors.append(Connector(**connector_data))

        return [
            connector for connector
            in connectors + custom_connectors + additional_used_connectors
            if connector.name in connector_names and connector.name not in self._IGNORED_CONNECTORS
        ]

    def _fetch_single_page_connections(self, token: str, url: str) -> tuple[list[_ConnectionWithConnectorName], str]:
        response_data = self._fetch_single_page(url, headers={
            'Authorization': f'Bearer {token}',
        })

        connections = [Connection(**connection) for connection in response_data.get('value', [])]
        connection_with_connector_name = [
            _ConnectionWithConnectorName(
                id=connection.id,
                name=connection.name,
                connector_name=connector_name,
            )
            for connection in connections
            if (connector_name := connection.id.split('/')[-3]) not in self._IGNORED_CONNECTORS
        ]
        next_page = response_data.get('nextLink', None)

        return connection_with_connector_name, next_page

    def _fetch_all_connections(self, token: str) -> list[_ConnectionWithConnectorName]:
        next_page_url = self._get_all_connections_url()
        connections: list[_ConnectionWithConnectorName] = []

        while next_page_url:
            connections_page, next_page_url = self._fetch_single_page_connections(
                token=token,
                url=next_page_url,
            )
            connections.extend(connections_page)
            self._resource_count += len(connections_page)

        return connections

    def _do_fetch_resource_data(self) -> list[ConnectorWithConnections]:
        token = self._token_manager.fetch_access_token(Requests.ENVIRONMENTS_SCOPE)
        connections = self._fetch_all_connections(token)
        used_connector_names = list({connection.connector_name for connection in connections})

        self._resource_count = 0 # Reset the resource count before fetching connectors
        connectors = self._fetch_connectors(
            token=token,
            connector_names=used_connector_names
        )

        connectors_with_connections = map_(connectors, lambda connector: ConnectorWithConnections(
            connector=connector,
            connections=filter_(connections, lambda connection: connection.connector_name == connector.name)
        ))

        # Return connectors sorted by the number of connections
        return sorted(connectors_with_connections, key=lambda x: len(x.connections), reverse=True)



import random

from power_platform_security_assessment.base_classes import ConnectorWithConnections
from power_platform_security_assessment.security_features.common import extract_environment_ids_from_connectors
from power_platform_security_assessment.security_features.connectors.model import ConnectorsAnalysisResult


class ConnectorsAnalyzerTextualReport:

    @staticmethod
    def _select_example_connector(connectors: list[ConnectorWithConnections]) -> ConnectorWithConnections:
        # Select a random connector with at least one connection instance
        return random.choice([c for c in connectors if len(c.connections) >= 1])

    def _generate_textual_report_for_type(self, connectors: list[ConnectorWithConnections], connector_type: str) -> str:
        connectors_count = len(connectors)
        if connectors_count == 0:
            return ""

        connections_count = sum(len(connector.connections) for connector in connectors)
        environments_count = len(extract_environment_ids_from_connectors(connectors))

        textual_report = (
            f'There {"is" if connectors_count == 1 else "are"} <b>{connectors_count} {connector_type}</b> connector{"" if connectors_count == 1 else "s"} '
            f'in {environments_count} environment{"" if environments_count == 1 else "s"}. '
            f'{"This" if connectors_count == 1 else "These"} connector{"" if connectors_count == 1 else "s"} '
            f'hold{"s" if connectors_count == 1 else ""} <b>{connections_count}</b> connection instance{"" if connections_count == 1 else "s"}.\n'
        )

        example_connector = self._select_example_connector(connectors)
        example_connector_connections_count = len(example_connector.connections)
        textual_report += (
            f'<br>For example, the connector <b>{example_connector.connector.properties.displayName}</b> is <b>{connector_type}</b>, and '
            f'<b>{example_connector_connections_count}</b> connection instance{"" if example_connector_connections_count == 1 else "s"} '
            f'{"is" if example_connector_connections_count == 1 else "are"} found in your organization.<br>'
        )

        return textual_report

    def generate_textual_report(self, connectors_info: ConnectorsAnalysisResult) -> str:
        deprecated_connectors_text = self._generate_textual_report_for_type(connectors_info.deprecated_connectors, "deprecated")
        untrusted_connectors_text = self._generate_textual_report_for_type(connectors_info.untrusted_connectors, "untrusted")
        return (
            f'{deprecated_connectors_text}\n'
            f'{untrusted_connectors_text}\n'
        )



from pydantic import BaseModel

from power_platform_security_assessment.base_classes import Application, CloudFlow, User


class UserResources(BaseModel):
    user: User
    apps: list[Application]
    flows: list[CloudFlow]


class Developers(BaseModel):
    guest_developers: list[UserResources]
    inactive_developers: list[UserResources]


class AppDevelopersReport(BaseModel):
    developers_info: Developers
    textual_report: str



from urllib.parse import urlencode

from power_platform_security_assessment.base_classes import CloudFlow
from power_platform_security_assessment.consts import Requests
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class CloudFlowsFetcher(BaseResourceFetcher):
    def __init__(self, env_id: str, token_manager: TokenManager):
        super().__init__(env_id, token_manager)

    def _get_request_url(self):
        params = {"api-version": "2016-11-01"}
        return f'https://api.flow.microsoft.com/providers/Microsoft.ProcessSimple/scopes/admin/environments/{self._env_id}/v2/flows?{urlencode(params)}'

    def _fetch_single_page_cloud_flows(self, token: str, url: str):
        response_data = self._fetch_single_page(url, headers={
            'Authorization': f'Bearer {token}',
            'x-ms-api-context-type': 'Admin',
            'Accept-Encoding': 'gzip,deflate,compress',
        })

        cloud_flows = [CloudFlow(**flow) for flow in response_data.get('value', [])]
        next_page = response_data.get('nextLink', None)

        return cloud_flows, next_page

    def _fetch_cloud_flows(self) -> list[CloudFlow]:
        token = self._token_manager.fetch_access_token(Requests.ENVIRONMENTS_SCOPE)
        next_page_url = self._get_request_url()
        all_cloud_flows = []

        while next_page_url:
            cloud_flows, next_page_url = self._fetch_single_page_cloud_flows(token=token, url=next_page_url)
            all_cloud_flows.extend(cloud_flows)
            self._resource_count += len(cloud_flows)

        return all_cloud_flows

    def _do_fetch_resource_data(self) -> list[CloudFlow]:
        return self._fetch_cloud_flows()



import time
from abc import abstractmethod

import requests
from pydash import take

from power_platform_security_assessment.base_classes import T, ResourceData
from power_platform_security_assessment.token_manager import TokenManager


class BaseResourceFetcher:
    _DEFAULT_MAX_RESOURCE_COUNT = 5000  # limit of the number of resources to fetch

    def __init__(self, env_id: str, token_manager: TokenManager):
        self._env_id = env_id
        self._token_manager = token_manager

        self._resource_count = 0
        self._max_resource_count = self._DEFAULT_MAX_RESOURCE_COUNT
        self._exceeded_max_resource_count = False

    def _retry(self, res: requests.Response, url: str, headers: dict, attempts: int):
        if attempts <= 0:
            print(f"Error: {res.status_code} while fetching {url}, response: {res.text}")
            raise Exception(f"Rate limit exceeded for {url}")

        retry_after = res.headers.get('Retry-After') or res.headers.get('retry-after')
        time.sleep(int(retry_after))
        return self._fetch_single_page(url, headers, attempts)

    def _fetch_single_page(self, url: str, headers: dict, attempts: int = 3):
        if self._resource_count >= self._max_resource_count:
            self._exceeded_max_resource_count = True
            return {}

        res = requests.get(url, headers=headers)
        if res.status_code == 429:  # Too Many Requests
            return self._retry(
                res=res,
                url=url,
                headers=headers,
                attempts=attempts - 1,
            )

        switcher = {
            200: res.json(),
            404: {},
        }

        if res.status_code not in switcher:
            res.raise_for_status()

        return switcher.get(res.status_code)

    def fetch_resource_data(self) -> ResourceData[T]:
        resource_data = self._do_fetch_resource_data()
        value = take(resource_data, self._max_resource_count)

        return ResourceData(
            value=value,
            count=len(value),
            all_resources_fetched=len(resource_data) == len(value)
                                  and not self._exceeded_max_resource_count,
        )

    @abstractmethod
    def _do_fetch_resource_data(self) -> list[T]:
        raise NotImplementedError



import random
from typing import Union

from pydash import find, flatten, chain

from power_platform_security_assessment.base_classes import Environment, Application, CloudFlow
from power_platform_security_assessment.security_features.app_developers.model import UserResources, Developers
from power_platform_security_assessment.security_features.common import (
    extract_environment_id, extract_user_domain, is_app_disabled, is_flow_disabled
)


class AppDeveloperTextualReport:
    def __init__(self, environments: list[Environment]):
        self._environments = environments

    @staticmethod
    def _select_example_user(user_resources: list[UserResources]) -> UserResources:
        # Select a random user with at least one app or flow
        return random.choice([u for u in user_resources if len(u.apps) + len(u.flows) >= 1])

    def _get_environment_names(self, apps: list[Union[Application, CloudFlow]]) -> list[str]:
        environment_ids = {extract_environment_id(app.id) for app in apps}
        return [
            find(self._environments, lambda env: env.name.lower() == env_id.lower()).properties.displayName
            for env_id in environment_ids
        ]

    def _generate_env_text(self, resources: list, resource_type: str):
        apps_envs = self._get_environment_names(flatten(resources))
        envs_text = f'<b>{"</b>, <b>".join(apps_envs)}</b> environment{"" if len(apps_envs) == 1 else "s"}'
        return f'<b>{len(resources)}</b> {resource_type}{"" if len(resources) == 1 else "s"} in the {envs_text}.'

    def _generate_developer_textual_report(self, user_resources: list[UserResources], developer_type: str) -> str:
        users_count = len(user_resources)
        apps_count = sum(len(developer.apps) for developer in user_resources)
        flows_count = sum(len(developer.flows) for developer in user_resources)

        disabled_apps_count = (
            chain(user_resources)
            .flat_map(lambda developer: developer.apps)
            .filter_(lambda app: is_app_disabled(app))
            .size()
            .value()
        )
        disabled_flows_count = (
            chain(user_resources)
            .flat_map(lambda developer: developer.flows)
            .filter_(lambda flow: is_flow_disabled(flow))
            .size()
            .value()
        )
        total_disabled_count = disabled_apps_count + disabled_flows_count
        total_active_count = apps_count + flows_count - total_disabled_count

        if users_count == 0 or apps_count + flows_count == 0:
            return ""

        textual_report = (
            f'There are <b>{apps_count + flows_count}</b> different applications and flows '
            f'owned by <b>{users_count}</b> different <b>{developer_type}</b> users. '
            f'<b>{total_disabled_count}</b> {"is" if total_disabled_count == 1 else "are"} disabled '
            f'and <b>{total_active_count}</b> {"is" if total_active_count == 1 else "are"} active.'
        )

        example_user = self._select_example_user(user_resources)
        textual_report += f'<br>For example, <b>{example_user.user.fullname}</b> from <b>{extract_user_domain(example_user.user)}</b> developed '

        if example_user.apps:
            textual_report += self._generate_env_text(example_user.apps, 'application')

        if example_user.flows:
            if example_user.apps:
                textual_report += ' and '  # Add "and" only if there are apps
            textual_report += self._generate_env_text(example_user.flows, 'flow')

        return textual_report + '<br>'

    def generate_textual_report(self, developers: Developers) -> str:
        guest_developers_textual_report = self._generate_developer_textual_report(developers.guest_developers, 'guest')
        inactive_developers_textual_report = self._generate_developer_textual_report(developers.inactive_developers, 'deleted')
        return (
            f'{guest_developers_textual_report}'
            f'{inactive_developers_textual_report}'
        )



import concurrent.futures

from power_platform_security_assessment.base_classes import (
    Environment, User, Application, CloudFlow, ConnectorWithConnections, DesktopFlow, ModelDrivenApp, ResourceData
)
from power_platform_security_assessment.consts import ComponentType
from power_platform_security_assessment.fetchers.applications_fetcher import ApplicationsFetcher
from power_platform_security_assessment.fetchers.cloud_flows_fetcher import CloudFlowsFetcher
from power_platform_security_assessment.fetchers.connections_fetcher import ConnectionsFetcher
from power_platform_security_assessment.fetchers.desktop_flows_fetcher import DesktopFlowsFetcher
from power_platform_security_assessment.fetchers.model_driven_apps_fetcher import ModelDrivenAppsFetcher
from power_platform_security_assessment.fetchers.users_fetcher import UsersFetcher
from power_platform_security_assessment.token_manager import TokenManager


class EnvironmentScanner:
    def __init__(self, environment: Environment, token_manager: TokenManager):
        self._environment = environment
        self._env_id = environment.id.split('/')[-1]
        self._token_manager = token_manager

    def _fetch_applications(self) -> ResourceData[Application]:
        app_fetcher = ApplicationsFetcher(
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return app_fetcher.fetch_resource_data()

    def _fetch_cloud_flows(self) -> ResourceData[CloudFlow]:
        cloud_flow_fetcher = CloudFlowsFetcher(
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return cloud_flow_fetcher.fetch_resource_data()

    def _fetch_desktop_flows(self) -> ResourceData[DesktopFlow]:
        if not self._environment.properties.linkedEnvironmentMetadata:
            return ResourceData()

        desktop_flow_fetcher = DesktopFlowsFetcher(
            instance_api_url=self._environment.properties.linkedEnvironmentMetadata.instanceApiUrl,
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return desktop_flow_fetcher.fetch_resource_data()

    def _fetch_model_driven_apps(self) -> ResourceData[ModelDrivenApp]:
        if not self._environment.properties.linkedEnvironmentMetadata:
            return ResourceData()

        model_driven_apps_fetcher = ModelDrivenAppsFetcher(
            instance_api_url=self._environment.properties.linkedEnvironmentMetadata.instanceApiUrl,
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return model_driven_apps_fetcher.fetch_resource_data()

    def _fetch_connections(self) -> ResourceData[ConnectorWithConnections]:
        connections_fetcher = ConnectionsFetcher(
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return connections_fetcher.fetch_resource_data()

    def _fetch_users(self) -> ResourceData[User]:
        if not self._environment.properties.linkedEnvironmentMetadata:
            return ResourceData()

        users_fetcher = UsersFetcher(
            instance_api_url=self._environment.properties.linkedEnvironmentMetadata.instanceApiUrl,
            env_id=self._env_id,
            token_manager=self._token_manager,
        )

        return users_fetcher.fetch_resource_data()

    def scan_environment(self):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = {ComponentType.ENVIRONMENT: self._environment, 'error': None}
            futures = {
                executor.submit(self._fetch_applications): ComponentType.APPLICATIONS,
                executor.submit(self._fetch_cloud_flows): ComponentType.CLOUD_FLOWS,
                executor.submit(self._fetch_desktop_flows): ComponentType.DESKTOP_FLOWS,
                executor.submit(self._fetch_model_driven_apps): ComponentType.MODEL_DRIVEN_APPS,
                executor.submit(self._fetch_connections): ComponentType.CONNECTIONS,
                executor.submit(self._fetch_users): ComponentType.USERS,
            }

            for future in concurrent.futures.as_completed(futures):
                data_type: ComponentType = futures[future]
                try:
                    results[data_type] = future.result()
                except Exception as e:
                    results[data_type] = None
                    results['error'] = e

            return results



from dateutil import parser

from power_platform_security_assessment.consts import ComponentType
from power_platform_security_assessment.security_features.common import get_application_owner_id, \
    get_cloud_flow_owner_id, get_desktop_flow_owner_id, get_model_driven_app_owner_id


def get_environment_developers_count(environment_results) -> int:
    component_mappings = {
        ComponentType.APPLICATIONS: get_application_owner_id,
        ComponentType.CLOUD_FLOWS: get_cloud_flow_owner_id,
        ComponentType.DESKTOP_FLOWS: get_desktop_flow_owner_id,
        ComponentType.MODEL_DRIVEN_APPS: get_model_driven_app_owner_id,
    }

    # Use a set to store unique developer IDs
    developers = {
        user_id
        for component_type, get_owner_id in component_mappings.items()
        for component in environment_results[component_type].value
        if (user_id := get_owner_id(component))
    }

    return len(developers)


def round_time_to_seconds(time: str):
    return parser.isoparse(time).strftime('%Y/%m/%d %H:%M:%S')


DEPRECATED_CONNECTOR_NAMES = [
    'Azure Log Analytics',
    'ConsenSys Ethereum',
    'Corda Blockchain',
    'D&B Optimizer',
    'Dynamic 365',
    'Elastic OCR',
    'Ethereum Blockchain',
    'Excel',
    'Language Tool',
    'LinkedIn',
    'Microsoft Dataverse (Legacy)',
    'Microsoft Translator',
    'Module Q',
    'Office 365 Video',
    'Outlook Tasks',
    'Pitney Bowes Data Validation',
    'Pitney Bowes Tax Calculator',
    'Projectwise Share',
    'Recorded Future',
    'SFTP',
    'Skype for Business Online',
    'Todoist',
    'Xpertdoc',
]



import concurrent.futures
from collections import Counter

import msal
from alive_progress import alive_bar
from pydash import flatten_deep, values

from power_platform_security_assessment.base_classes import (
    Environment, User, ConnectorWithConnections, Application, ResourceData, CloudFlow, DesktopFlow, ModelDrivenApp
)
from power_platform_security_assessment.consts import Requests, ResponseKeys, ComponentType
from power_platform_security_assessment.environment_scanner import EnvironmentScanner
from power_platform_security_assessment.fetchers.environments_fetcher import EnvironmentsFetcher
from power_platform_security_assessment.report_builder.report_builder import ReportBuilder
from power_platform_security_assessment.security_features.app_developers.app_developer_analyzer import \
    AppDeveloperAnalyzer
from power_platform_security_assessment.security_features.bypass_consent.bypass_consent_analyzer import \
    BypassConsentAnalyzer
from power_platform_security_assessment.security_features.connectors.connectors_analyzer import ConnectorsAnalyzer
from power_platform_security_assessment.token_manager import TokenManager


class SecurityAssessmentTool:
    def __init__(self):
        self._access_token = None
        self._refresh_token = None
        self._client_id = None

    def _create_token(self):
        app = msal.PublicClientApplication('9cee029c-6210-4654-90bb-17e6e9d36617', authority=Requests.AUTHORITY)
        result = app.acquire_token_interactive(scopes=Requests.ENVIRONMENTS_SCOPE)

        if ResponseKeys.ACCESS_TOKEN in result:
            self._access_token = result[ResponseKeys.ACCESS_TOKEN]
            self._refresh_token = result.get("refresh_token")
            self._client_id = result.get("id_token_claims").get("aud")
        else:
            raise Exception("Failed to acquire token: %s" % result.get("error_description"))

    @staticmethod
    def _scan_environment(environment: Environment, token_manager: TokenManager):
        env_scanner = EnvironmentScanner(
            environment=environment,
            token_manager=token_manager,
        )
        return env_scanner.scan_environment()

    @staticmethod
    def _display_environment_results(environments_results, failed_environments):
        print(
            f'{"Environment":<44} {"Canvas Apps":<15} {"Cloud Flows":<15} {"Desktop Flows":<15} {"Model-Driven Apps":<18} {"Total":<15}')

        results: list[tuple[str, ResourceData, ResourceData, ResourceData, ResourceData, int]] = []
        for environment_results in environments_results:
            environment: Environment = environment_results[ComponentType.ENVIRONMENT]

            applications_result: ResourceData[Application] = environment_results[ComponentType.APPLICATIONS]
            cloud_flows_result: ResourceData[CloudFlow] = environment_results[ComponentType.CLOUD_FLOWS]
            desktop_flows_result: ResourceData[DesktopFlow] = environment_results[ComponentType.DESKTOP_FLOWS]
            model_driven_apps_result: ResourceData[ModelDrivenApp] = environment_results[
                ComponentType.MODEL_DRIVEN_APPS]

            results.append((
                environment.properties.displayName,
                applications_result,
                cloud_flows_result,
                desktop_flows_result,
                model_driven_apps_result,
                applications_result.count + cloud_flows_result.count + desktop_flows_result.count + model_driven_apps_result.count
            ))

        # sort by total
        results = sorted(results, key=lambda x: x[5], reverse=True)

        # display results
        for result in results:
            total_has_plus = any(
                not result[i].all_resources_fetched
                for i in range(1, 5)
            )
            print(
                f'{result[0]:<44} '
                f'{str(result[1].count) + ("+" if not result[1].all_resources_fetched else ""):<15} '
                f'{str(result[2].count) + ("+" if not result[2].all_resources_fetched else ""):<15} '
                f'{str(result[3].count) + ("+" if not result[3].all_resources_fetched else ""):<15} '
                f'{str(result[4].count) + ("+" if not result[4].all_resources_fetched else ""):<18} '
                f'{str(result[5]) + ("+" if total_has_plus else ""):<15} '
            )

        if failed_environments:
            print()
            print('Environments failed to scan - Insufficient user permissions:')
            for failed_environment in failed_environments:
                print(f'{failed_environment[ComponentType.ENVIRONMENT].properties.displayName}')

        print()

    @staticmethod
    def _handle_environment_users(environments_results) -> tuple[list[User], bool]:
        all_fetched = True
        users_list: list[User] = []
        for environment_results in environments_results:
            all_fetched = all_fetched and environment_results[ComponentType.USERS].all_resources_fetched
            environment_users: list[User] = environment_results[ComponentType.USERS].value
            existing_user_ids = {u.azureactivedirectoryobjectid for u in users_list}
            users_list.extend(
                user for user in environment_users
                if user.azureactivedirectoryobjectid and user.azureactivedirectoryobjectid not in existing_user_ids
            )

        return users_list, all_fetched

    @staticmethod
    def _handle_connector_connections(environments_results):
        # Dictionary to map connector names to their respective ConnectorWithConnections objects
        connector_mapping: dict[str, ConnectorWithConnections] = {}

        for environment_results in environments_results:
            for connector_with_connections in environment_results[ComponentType.CONNECTIONS].value:
                connector_name = connector_with_connections.connector.name
                if connector_name in connector_mapping:
                    connector_mapping[connector_name].connections.extend(connector_with_connections.connections)
                else:
                    connector_mapping[connector_name] = connector_with_connections

        # Convert the dictionary back to a list
        all_connector_connections = values(connector_mapping)

        # Return connectors sorted by the number of connections
        return sorted(all_connector_connections, key=lambda x: len(x.connections), reverse=True)

    @staticmethod
    def _display_users(users_list: list[User], all_users_fetched: bool):
        print(f'{"Total Users":<22}')
        print(f'{len(users_list)}' + ("+" if not all_users_fetched else ""))

        user_types = Counter()
        azure_states = Counter()

        for user in users_list:
            user_types['guest' if '#EXT#' in (user.domainname or '') else 'internal'] += 1
            azure_states[user.azurestate] += 1

        print(f'{"Internal Users":<22} {"Guest Users":<22}')
        print(f'{user_types["internal"]:<22} {user_types["guest"]:<22}')

        print(f'{"Active (0)":<22} {"AD Soft Delete (1)":<22} {"AD Hard Delete (2)":<22}')
        print(f'{azure_states[0]:<22} {azure_states[1]:<22} {azure_states[2]:<22}')
        print()

    @staticmethod
    def _display_connections(all_connector_connections: list[ConnectorWithConnections]):
        print(f'{"Connector":<22} {"Publisher":<22} {"Number of Connections":<22}')
        for connector_with_connections in all_connector_connections[:3]:
            connector = connector_with_connections.connector
            connections_count = len(connector_with_connections.connections)
            print(f'{connector.name:<22} {connector.properties.publisher:<22} {connections_count:<22}')

        print()

    @staticmethod
    def _display_app_developers(all_applications, all_cloud_flows, users_list, environments):
        app_developer_analyzer = AppDeveloperAnalyzer(all_applications, all_cloud_flows, users_list, environments)
        app_developers_result = app_developer_analyzer.analyze()
        return app_developers_result.textual_report

    @staticmethod
    def _display_connector_issues(all_connector_connections: list[ConnectorWithConnections]):
        connectors_analyzer = ConnectorsAnalyzer(all_connector_connections)
        result = connectors_analyzer.analyze()
        return result.textual_report

    @staticmethod
    def _display_bypass_consent(all_applications: list[Application]):
        bypass_consent_analyzer = BypassConsentAnalyzer(all_applications)
        bypass_consent_result = bypass_consent_analyzer.analyze()
        return bypass_consent_result.textual_report

    def _handle_results(self, environments_results: list, failed_environments: list, environments: list[Environment], total_envs: int):
        (all_applications, all_cloud_flows, all_connector_connections, all_desktop_flows, all_model_driven_apps,
         all_users_list, all_users_fetched) = self.fetch_resources(environments_results)

        app_developers_report = self._display_app_developers(all_applications, all_cloud_flows, all_users_list,
                                                             environments)
        connector_issues_report = self._display_connector_issues(all_connector_connections)
        bypass_consent_report = self._display_bypass_consent(all_applications)
        self._display_environment_results(environments_results, failed_environments)
        self._display_users(all_users_list, all_users_fetched)
        self._display_connections(all_connector_connections)

        report_builder = ReportBuilder(all_applications, all_cloud_flows, all_desktop_flows, all_model_driven_apps,
                                       all_users_list, all_connector_connections, environments_results,
                                       failed_environments, environments, total_envs, all_users_fetched)

        report_builder.build_report(app_developers_report, connector_issues_report, bypass_consent_report)

    def fetch_resources(self, environments_results):
        all_users_list, all_users_fetched = self._handle_environment_users(environments_results)
        all_connector_connections = self._handle_connector_connections(environments_results)
        all_applications = flatten_deep(
            [env_results[ComponentType.APPLICATIONS].value for env_results in environments_results])
        all_cloud_flows = flatten_deep(
            [env_results[ComponentType.CLOUD_FLOWS].value for env_results in environments_results])
        all_desktop_flows = flatten_deep(
            [env_results[ComponentType.DESKTOP_FLOWS].value for env_results in environments_results])
        all_model_driven_apps = flatten_deep(
            [env_results[ComponentType.MODEL_DRIVEN_APPS].value for env_results in environments_results])
        return all_applications, all_cloud_flows, all_connector_connections, all_desktop_flows, all_model_driven_apps, all_users_list, all_users_fetched

    def run_security_assessment(self):
        print('Started Scanning Environments...')
        self._create_token()
        environments, total_envs = EnvironmentsFetcher().fetch_environments(self._access_token)
        token_manager = TokenManager(self._client_id, self._refresh_token)
        environments_results = []
        failed_environments = []

        with alive_bar(len(environments), bar='blocks') as bar:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [executor.submit(self._scan_environment, environment, token_manager) for environment in
                           environments]
                for future in concurrent.futures.as_completed(futures):
                    try:
                        result = future.result()
                        if not result['error']:
                            environments_results.append(result)
                        else:
                            failed_environments.append(result)
                    except Exception as e:
                        print(f"An error occurred during environment scanning: {e}")
                    finally:
                        bar()

        self._handle_results(environments_results, failed_environments, environments, total_envs)


def main():
    security_assessment_tool = SecurityAssessmentTool()
    security_assessment_tool.run_security_assessment()


if __name__ == '__main__':
    main()



from urllib.parse import urlencode

from power_platform_security_assessment.base_classes import User
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class UsersFetcher(BaseResourceFetcher):
    def __init__(self, instance_api_url: str, env_id: str, token_manager: TokenManager):
        self._instance_api_url = instance_api_url
        super().__init__(env_id, token_manager)
        self._max_resource_count = 50_000

    def _get_request_url(self):
        params = {
            "$select": "fullname,domainname,isdisabled,azurestate,azureactivedirectoryobjectid",
            "$filter": "azureactivedirectoryobjectid ne null and (accessmode eq 0 or accessmode eq 1 or accessmode eq 2)",
        }
        return f'{self._instance_api_url}/api/data/v9.2/systemusers?{urlencode(params)}'

    def _fetch_single_users_page(self, token: str, url: str):
        response_data = self._fetch_single_page(url, {
            'Authorization': f'Bearer {token}',
        })

        users = [User(**user) for user in response_data.get('value', [])]
        next_page = response_data.get('@odata.nextLink', None)

        return users, next_page

    def _fetch_users(self) -> list[User]:
        token = self._token_manager.fetch_access_token(scope=[f'{self._instance_api_url}/.default'])
        next_page_url = self._get_request_url()
        users_list: list[User] = []

        while next_page_url:
            users, next_page_url = self._fetch_single_users_page(token=token, url=next_page_url)
            users_list.extend(users)
            self._resource_count += len(users)

        return users_list

    def _do_fetch_resource_data(self) -> list[User]:
        return self._fetch_users()



from enum import Enum


class Requests:
    AUTHORITY = "https://login.microsoftonline.com/common"
    CLIENT_ID = "23d8f6bd-1eb0-4cc2-a08c-7bf525c67bcd"
    ENVIRONMENTS_SCOPE = ['https://service.powerapps.com/.default']
    APPLICATIONS_SCOPE = ['https://api.powerplatform.com/.default']


class ResponseKeys:
    ACCESS_TOKEN = "access_token"


class ComponentType(Enum):
    ENVIRONMENT = 'environment'
    APPLICATIONS = 'applications'
    CLOUD_FLOWS = 'cloud_flows'
    DESKTOP_FLOWS = 'desktop_flows'
    MODEL_DRIVEN_APPS = 'model_driven_apps'
    CONNECTIONS = 'connections'
    USERS = 'users'
